from void import (
    M,
    duo1,
    leader,
)


if leader is None:
    leader = duo1.radar(
        M.at.player, M.at.any, M.at.any,
        M.at.distance, M.at.asc)
elif leader.health <= 0:
    leader = None
else:
    M.unit.bind(M.at.nova)
    M.unit.approach(leader.x, leader.y, 5)
    M.unit.targetPosition(leader.shootX, leader.shootY, leader.shooting)



def parse_programs(pool):
    progs = []

    code = []
    output = []
    reading_code = True

    def flush():
        nonlocal reading_code
        if code or output:
            progs.append((
                '\n'.join(code),
                '\n'.join(output),
            ))
        code.clear()
        output.clear()
        reading_code = True

    for line in pool.strip().splitlines():
        if line.startswith('==='):
            flush()
        elif line.startswith('---'):
            reading_code = False
        else:
            if reading_code:
                code.append(line)
            else:
                output.append(line)
    flush()

    return progs



from void import (
    M,
    message1, sorter1, vault1, container1,
    last_ore,
)

container = vault1 or container1
if not container:
    M.print('No linked container')
    message1.printFlush()
    M.exit()
core_x = container.x
core_y = container.y

if not sorter1.config:
    M.print('Choose ore')
    message1.printFlush()
    M.exit()

M.unit.bind(M.at.mega)

if last_ore != sorter1.config:
    if M.unit.within(core_x, core_y, 5):
        found, ore_x, ore_y = M.locate.ore(sorter1.config)
        if found:
            last_ore = sorter1.config
        else:
            M.print('Can\'t locate ore')
            message1.printFlush()
        M.exit()
    else:
        M.print('Locating ore...')
        message1.printFlush()
        M.unit.approach(core_x, core_y, 5)
    M.exit()

M.print('Gathering...')
message1.printFlush()

dst_core = M.dst(M.unit.x - core_x, M.unit.y - core_y)
dst_ore = M.dst(M.unit.x - ore_x, M.unit.y - ore_y)

closer_to_core = dst_core < dst_ore

items = M.unit.totalItems
cap = M.unit.itemCapacity

full = items == cap
some = 0 < items < cap

if full or (some and closer_to_core):
    if M.unit.within(core_x, core_y, 5):
        M.unit.itemDrop(container, M.unit.totalItems)
    else:
        M.unit.approach(core_x, core_y, 5)
else:
    if M.unit.within(ore_x, ore_y, 5):
        M.unit.mine(ore_x, ore_y)
    else:
        M.unit.approach(ore_x, ore_y, 5)



from void import (
    M,
    switch1, display1,
    setup, time,
)

# large display is 176 x 176
dsize = 176

M.draw.clear(0, 0, 0)
M.draw.stroke(5)
M.draw.color(255, 0, 0, 255)


def posvel(pos, vel):
    pos += vel
    if pos < 0:
        pos = 0
        vel = M.abs(vel)
    if pos > dsize:
        pos = dsize
        vel = -M.abs(vel)
    return pos, vel


def randvel():
    variation = 10
    return M.rand(variation * 2) - variation


if setup is None or switch1.enabled:
    t1x, t1y = M.rand(dsize), M.rand(dsize)
    t2x, t2y = M.rand(dsize), M.rand(dsize)
    t3x, t3y = M.rand(dsize), M.rand(dsize)
    t1vx, t1vy = randvel(), randvel()
    t2vx, t2vy = randvel(), randvel()
    t3vx, t3vy = randvel(), randvel()
    setup = True
else:
    t1x, t1vx = posvel(t1x, t1vx)
    t1y, t1vy = posvel(t1y, t1vy)
    t2x, t2vx = posvel(t2x, t2vx)
    t2y, t2vy = posvel(t2y, t2vy)
    t3x, t3vx = posvel(t3x, t3vx)
    t3y, t3vy = posvel(t3y, t3vy)


M.draw.line(t1x, t1y, t2x, t2y)
M.draw.line(t2x, t2y, t3x, t3y)
M.draw.line(t3x, t3y, t1x, t1y)


time += 10
if time > 360:
    time %= 360


M.draw.color(255, 255, 255, 255)
M.draw.image(t1x, t1y, M.at.lead, 30 + M.sin(time) * 10, time)
M.draw.image(t2x, t2y, M.at.metaglass, 30 + M.sin(time + 120) * 10, time)
M.draw.image(t3x, t3y, M.at.thorium, 30 + M.sin(time + 240) * 10, time)


display1.drawFlush()



from .mast import Jump, Label


_opts = []


def reg(fn):
    _opts.append(fn)
    return fn


@reg
def remove_unnecessary_jumps(instructions):
    remove = set()
    for index, (a, b) in enumerate(zip(instructions, instructions[1:])):
        if isinstance(a, Jump) and isinstance(b, Label):
            if a.label is b:
                remove.add(index)
    if not remove:
        return None
    return [ins for i, ins in enumerate(instructions) if i not in remove]


def optimize(instructions):
    applied = 1
    while applied > 0:
        applied = 0
        for o in _opts:
            ni = o(instructions)
            if ni is not None:
                instructions = ni
                applied += 1
    return instructions



from typing import Optional, Union


class MConst:
    pass


class MObject:
    pass


Value = Union[float, MObject, MConst]
ColorComponent = float  # 0..255
Boolean = float  # 0 or 1
Angle = float  # 0..360
Integer = float  # number will be casted to int


def print(*args: Value) -> None:
    pass


def __print_flush(target: MObject) -> None:
    # this function is available only as print.flush
    pass


print.flush = __print_flush


def exit() -> None:
    pass


def min(a: float, b: float) -> float:
    pass


def max(a: float, b: float) -> float:
    pass


def atan2(x: float, y: float) -> Angle:
    pass


def dst(x: float, y: float) -> float:
    "Distance from (0, 0) to (x, y)"


def noise(x: float, y: float) -> float:
    "Simplex noise"


def abs(v: float) -> float:
    pass


def log(v: float) -> float:
    pass


def log10(v: float) -> float:
    pass


def sin(v: Angle) -> float:
    pass


def cos(v: Angle) -> float:
    pass


def tan(v: Angle) -> float:
    pass


def floor(v: float) -> Integer:
    pass


def ceil(v: float) -> Integer:
    pass


def sqrt(v: float) -> float:
    pass


def rand(range_: float) -> float:
    pass


def GetLink(index: int) -> Optional[MObject]:
    pass


def Radar(
        unit: MObject,
        target1: MConst, target2: MConst, target3: MConst,
        sort_type: MConst,
        sort_dir: MConst) -> Optional[MObject]:
    pass


def Sensor(unit: MObject, prop: MConst) -> Value:
    pass


class Draw:
    @staticmethod
    def clear(
            r: ColorComponent,
            g: ColorComponent,
            b: ColorComponent) -> None:
        pass

    @staticmethod
    def color(
            r: ColorComponent,
            g: ColorComponent,
            b: ColorComponent,
            a: ColorComponent) -> None:
        pass

    @staticmethod
    def stroke(width: float) -> None:
        pass

    @staticmethod
    def line(x: float, y: float, x2: float, y2: float) -> None:
        pass

    @staticmethod
    def rect(x: float, y: float, width: float, height: float) -> None:
        pass

    @staticmethod
    def lineRect(x: float, y: float, width: float, height: float) -> None:
        pass

    @staticmethod
    def poly(
            x: float, y: float,
            sides: float, radius: float, rotation: float) -> None:
        pass

    @staticmethod
    def linePoly(
            x: float, y: float,
            sides: float, radius: float, rotation: float) -> None:
        pass

    @staticmethod
    def triangle(
            x: float, y: float,
            x2: float, y2: float,
            x3: float, y3: float) -> None:
        pass

    @staticmethod
    def image(x: float, y: float, image: MConst, size: float, rotation: float):
        pass

    @staticmethod
    def flush(target: MObject) -> None:
        pass


class Control:
    @staticmethod
    def setEnabled(unit: MObject, is_enabled: Boolean) -> None:
        pass

    @staticmethod
    def targetPosition(unit: MObject, x: float, y: float, shoot: bool) -> None:
        pass

    @staticmethod
    def targetObject(unit: MObject, target: MObject, shoot: bool) -> None:
        pass

    @staticmethod
    def configure(unit: MObject, value: MConst) -> None:
        pass


# List of constants
# https://github.com/Anuken/Mindustry/blob/master/core/src/mindustry/logic/GlobalConstants.java#L15


# https://github.com/Anuken/Mindustry/blob/master/core/src/mindustry/content/Items.java
class Material:
    copper: MConst
    lead: MConst
    metaglass: MConst
    graphite: MConst
    sand: MConst
    coal: MConst
    titanium: MConst
    thorium: MConst
    scrap: MConst
    silicon: MConst
    plastanium: MConst
    phase_fab: MConst
    surge_alloy: MConst
    spore_pod: MConst
    blast_compound: MConst
    pyratite: MConst


# https://github.com/Anuken/Mindustry/blob/master/core/src/mindustry/content/Liquids.java
class Liquid:
    water: MConst
    slag: MConst
    oil: MConst
    cryofluid: MConst


# https://github.com/Anuken/Mindustry/blob/master/core/src/mindustry/logic/LAccess.java
class Property:
    totalItems: MConst
    firstItem: MConst
    totalLiquids: MConst
    totalPower: MConst
    itemCapacity: MConst
    liquidCapacity: MConst
    powerCapacity: MConst
    powerNetStored: MConst
    powerNetCapacity: MConst
    powerNetIn: MConst
    powerNetOut: MConst
    ammo: MConst
    ammoCapacity: MConst
    health: MConst
    maxHealth: MConst
    heat: MConst
    efficiency: MConst
    rotation: MConst
    x: MConst
    y: MConst
    shootX: MConst
    shootY: MConst
    shooting: MConst
    mineX: MConst
    mineY: MConst
    mining: MConst
    team: MConst
    type: MConst
    flag: MConst
    controlled: MConst
    commanded: MConst
    name: MConst
    config: MConst
    payloadCount: MConst
    payloadType: MConst

    enabled: MConst
    configure: MConst


# https://github.com/Anuken/Mindustry/blob/master/core/src/mindustry/logic/RadarSort.java
class Sort:
    asc: MConst
    desc: MConst

    distance: MConst
    health: MConst
    shield: MConst
    armor: MConst
    maxHealth: MConst


# https://github.com/Anuken/Mindustry/blob/master/core/src/mindustry/logic/RadarTarget.java
class Target:
    any: MConst
    enemy: MConst
    ally: MConst
    player: MConst
    attacker: MConst
    flying: MConst
    boss: MConst
    ground: MConst


# https://github.com/Anuken/Mindustry/blob/master/core/src/mindustry/content/UnitTypes.java
class UnitType:
    dagger: MConst
    mace: MConst
    fortress: MConst
    scepter: MConst
    reign: MConst
    nova: MConst
    pulsar: MConst
    quasar: MConst
    vela: MConst
    corvus: MConst
    crawler: MConst
    atrax: MConst
    spiroct: MConst
    arkyid: MConst
    toxopid: MConst
    flare: MConst
    horizon: MConst
    zenith: MConst
    antumbra: MConst
    eclipse: MConst
    mono: MConst
    poly: MConst
    mega: MConst
    quad: MConst
    oct: MConst
    risso: MConst
    minke: MConst
    bryde: MConst
    sei: MConst
    omura: MConst
    alpha: MConst
    beta: MConst
    gamma: MConst
    block: MConst


# https://github.com/Anuken/Mindustry/blob/master/core/src/mindustry/world/meta/BlockFlag.java
class BlockFlag:
    core: MConst
    storage: MConst
    generator: MConst
    turret: MConst
    factory: MConst
    repair: MConst
    rally: MConst
    battery: MConst
    resupply: MConst
    reactor: MConst
    unitModifier: MConst
    extinguisher: MConst


# https://github.com/Anuken/Mindustry/blob/master/core/src/mindustry/content/Blocks.java
class Block:
    # TODO: put here full list of blocks

    air: MConst
    solid: MConst



from dataclasses import dataclass
from typing import Any, List, Union


@dataclass
class Literal:
    value: Any  # any scalar

    def dump(self):
        c = self.value
        if c is None:
            return 'null'
        elif c is False:
            return 'false'
        elif c is True:
            return 'true'
        elif isinstance(c, (int, float)):
            return str(c)
        elif isinstance(c, str):
            if '"' in c:
                raise ValueError(
                    f'Quotes are not allowed in literal strings {c!r}')
            return '"' + c.replace('\n', r'\n') + '"'
        else:
            raise ValueError(f'Unsupported constant {c!r}')


@dataclass
class Name:
    # if None must be named automatically
    name: str = None


Value = Union[Literal, Name]
LValue = Union[Name]


@dataclass
class ProcedureCall:
    op: str
    args: List[Value]


@dataclass
class FunctionCall:
    op: str
    args: List[Value]
    result: LValue


@dataclass
class Label:
    pass


@dataclass
class Jump:
    label: Label
    op: str  # condition
    args: List[Value]


Instruction = Union[ProcedureCall, FunctionCall, Label, Jump]


def filter_and_index_labels(instructions):
    labels = {}
    program = []
    for ins in instructions:
        if isinstance(ins, Label):
            labels[id(ins)] = len(program)
        else:
            program.append(ins)
    return program, labels


def create_dump_session(resolve_jump):
    from itertools import count

    c = count(start=1)

    def dump_argument(arg: Value):
        if isinstance(arg, Literal):
            return arg.dump()
        elif isinstance(arg, Name):
            if arg.name is None:
                arg.name = f'_r{next(c)}'
            return arg.name
        raise ValueError

    def dump_instruction(ins: Instruction):
        if isinstance(ins, ProcedureCall):
            return [
                ins.op,
                *map(dump_argument, ins.args),
            ]
        elif isinstance(ins, FunctionCall):
            return [
                ins.op,
                dump_argument(ins.result),
                *map(dump_argument, ins.args),
            ]
        elif isinstance(ins, Jump):
            return [
                'jump',
                str(resolve_jump(ins.label)),
                ins.op,
                *map(dump_argument, ins.args),
            ]
        raise ValueError

    return dump_instruction


def dump(instructions):
    instructions, labels = filter_and_index_labels(instructions)

    max_jump = -1

    def resolve_jump(label: Label) -> int:
        nonlocal max_jump
        value = labels[id(label)]
        if value > max_jump:
            max_jump = value
        return value

    result = []
    dump_instruction = create_dump_session(resolve_jump)
    for ins in instructions:
        result.append(' '.join(dump_instruction(ins)))

    if max_jump >= len(result):
        assert max_jump == len(result)
        result.append('end')

    return result



import argparse
import ast
import sys

from . import mind, mast, utils, optimizer


parser = argparse.ArgumentParser()
parser.add_argument(
    'file', type=argparse.FileType('r'))
parser.add_argument(
    '-o', '--output', type=argparse.FileType('w'), default=sys.stdout)
parser.add_argument('-c', '--clip', action='store_true')


def py_to_mind(code):
    program = []
    global_scope = utils.Scope(None)
    for stmt in ast.parse(code).body:
        _, lines = mind.transform_expr(stmt, global_scope)
        program.extend(lines)
    program = optimizer.optimize(program)
    return '\n'.join(mast.dump(program)) + '\n'


def main(in_file, out_file, use_clip):
    code = py_to_mind(in_file.read())
    if use_clip:
        try:
            import pyperclip
        except ImportError:
            print('Install pyperclip to use --clip option', file=sys.stderr)
            exit(1)
        pyperclip.copy(code)
        n = code.count("\n") + 1
        print(f'Copied {n} lines to clipboard')
    else:
        out_file.write(code)


def run():
    args = parser.parse_args()
    main(args.file, args.output, args.clip)



from copy import deepcopy

import pytest

from minpiler import emu, mparse, test_utils


@pytest.mark.parametrize('code,state,result', [
    ('op add r a 2', {}, '2'),
    ('op add r a 2', {'a': 2}, '4'),
    ('op add r a 2', {'a': -1}, '1'),
    ('set r 3', {}, '3'),
    ('set r "test"', {}, 'test'),
    ('read r cell1 8', {'cell1': {8: 9.}}, '9'),
    ('read r cell1 400', {'cell1': {8: 9.}}, '0'),
    ('read r cell56 4', {}, '0'),
    ('op not r 0', {}, '-1'),
    ('op not r 2', {}, '-3'),
    ('op or r 4 2', {}, '6'),
    ('op and r 7 3', {}, '3'),
    ('op land r 7 3', {}, '1'),
    ('op sub r 7 xxx', {}, '7'),
    ('op sub r 7 null', {}, '7'),
    ('op sub r 7 true', {}, '6'),
    ('op div r 7 0', {}, '0'),
    ('op idiv r 7 0', {}, '0'),
    ('op pow r 2 2', {}, '4'),
    # ('op abs r -7', {}, '7'),
])
def test_functions(code, state, result):
    ast = mparse.parse(f"{code}\nprint r")
    assert emu.execute(ast, deepcopy(state)) == result


PROGRAMS = """
print "hello"
--------------
hello
==============
set c true
jump 3 equal c true
print "impossible"
print "end"
--------------
end
==============
set c 8
set index 10
write c cell1 index
read v cell1 index
print v
--------------
8
==============
print 1
end
print 2
--------------
1
==============
"""


@pytest.mark.parametrize('code,result', test_utils.parse_programs(PROGRAMS))
def test_programs(code, result):
    ast = mparse.parse(code)
    assert emu.execute(ast, {
        'cell1': {},
    }) == result



from void import (
    M,
    unloader1, unloader2, unloader3, unloader4, unloader5, unloader6,
    container1, container2, container3, container4, container5, container6,
    container7,
    conveyor1, conveyor2, conveyor3,
)


def balance(container, unloader):
    lead = container.lead / 50
    sand = container.sand / 50
    coal = container.coal / 25

    if M.min(M.min(sand, coal), lead) >= 1:
        unloader.configure(M.at.spore_pod)
        return

    if lead < M.min(sand, coal):
        unloader.configure(M.at.lead)
    elif sand < coal:
        unloader.configure(M.at.sand)
    else:
        unloader.configure(M.at.coal)


balance(container1, unloader1)
balance(container2, unloader2)
balance(container3, unloader3)
balance(container4, unloader4)
balance(container5, unloader5)
balance(container6, unloader6)

conveyor1.setEnabled(container7.sand < 200)
conveyor2.setEnabled(container7.lead < 200)
conveyor3.setEnabled(container7.coal < 200)



from void import (
    M,
    unloader1, unloader2, unloader3,
)


# loads 1 plastanium conveyor using 3 unloaders
# ratio is for surge alloy


period = 10_000  # milliseconds
parts = 4
t = (M.at.time % period) / period * 4

unloader1.configure(M.at.lead)
unloader2.configure(M.at.silicon if t < 1 else M.at.copper)
unloader3.configure(M.at.silicon if t > 2 else M.at.titanium)



import ast
import sys
from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import Any, Callable

from . import mast, utils


_PY = (sys.version_info.major, sys.version_info.minor)


def _get_ast_slice(node):
    if _PY >= (3, 9):
        return node.slice
    else:
        assert isinstance(node.slice, ast.Index)
        return node.slice.value


BIN_OP_MAP = {
    ast.Add: 'add',
    ast.Sub: 'sub',
    ast.Mult: 'mul',
    ast.Div: 'div',
    ast.FloorDiv: 'idiv',
    ast.Mod: 'mod',
    ast.Pow: 'pow',
    ast.LShift: 'shl',
    ast.RShift: 'shr',
    ast.BitOr: 'or',
    ast.BitXor: 'xor',
    ast.BitAnd: 'and',
    # MatMult
}

COND_OP_MAP = {
    ast.Eq: 'equal',
    ast.NotEq: 'notEqual',
    ast.Lt: 'lessThan',
    ast.LtE: 'lessThanEq',
    ast.Gt: 'greaterThan',
    ast.GtE: 'greaterThanEq',
    ast.Is: 'equal',
}

BOOL_OP_MAP = {  # op, shortcut_condition
    ast.And: ('land', 'equal'),
    ast.Or: ('or', 'notEqual'),
}

UNARY_OP_MAP = {
    ast.Invert: lambda val, result: mast.FunctionCall(
        'op not', [val], result),
    ast.Not: lambda val, result: mast.FunctionCall(
        'op equal', [mast.Literal(0), val], result),
    ast.UAdd: lambda val, result: mast.FunctionCall(
        'op add', [mast.Literal(0), val], result),
    ast.USub: lambda val, result: mast.FunctionCall(
        'op sub', [mast.Literal(0), val], result),
}
RESERVED_NAMES = {
    'print',
    'exit',
    'min',
    'max',
    'atan2',
    'dst',
    'noise',
    'abs',
    'log',
    'log10',
    'sin',
    'cos',
    'tan',
    'floor',
    'ceil',
    'sqrt',
    'rand',
    'print',

    'GetLink',
    'Draw',
    'Control',
    'Radar',
    'Sensor',

    'Material',
    'Liquid',
    'Property',
    'Sort',
    'Target',
    'UnitType',
    'BuildingType',
    'Building',
}


def get_type_map(map, item, desc):
    if type(item) not in map:
        raise ValueError(f'Unsupported {desc} {item}')
    return map[type(item)]


@dataclass
class BaseExpressionHandler:
    expr: Any
    trec: Callable = field(repr=False)
    scope: utils.Scope = field(repr=False)
    pre: list = field(default_factory=list)
    resmap: dict = field(default_factory=dict)

    # AST_CLASS = ast.Xxx

    def dev_dump(self):
        print(ast.dump(self.expr))

    def get_results(self):
        return [self.resmap[i] for i in range(len(self.resmap))]

    def run_trec(self, expr):
        retvals, pre = self.trec(expr, self.scope)
        self.pre.extend(pre)
        return retvals

    def run_trec_single(self, expr):
        retvals = self.run_trec(expr)
        if len(retvals) < 1:
            return mast.Literal(None)
        return retvals[0]

    def proc(self, name, *args):
        self.pre.append(mast.ProcedureCall(name, args))

    def jump(self, label, op, *args):
        self.pre.append(mast.Jump(label, op, args))

    @contextmanager
    def sub_scope(self):
        self.scope = utils.Scope(self.scope)
        try:
            yield
        finally:
            self.scope = self.scope._parent_scope

    def handle(self):
        raise NotImplementedError


class ConstantHandler(BaseExpressionHandler):
    AST_CLASS = ast.Constant

    def handle(self):
        self.resmap[0] = mast.Literal(self.expr.value)


class NameHandler(BaseExpressionHandler):
    AST_CLASS = ast.Name

    def handle(self):
        if self.expr.id in self.scope:
            self.resmap[0] = self.scope[self.expr.id]
            return
        if self.expr.id in RESERVED_NAMES:
            raise ValueError(f'The name {self.expr.id} is reserved')
        self.resmap[0] = mast.Name(self.expr.id)


class TupleHandler(BaseExpressionHandler):
    AST_CLASS = ast.Tuple

    def handle(self):
        for index, value in enumerate(self.expr.elts):
            self.resmap[index] = self.run_trec_single(value)


class SubscriptHandler(BaseExpressionHandler):
    AST_CLASS = ast.Subscript

    def handle(self):
        # memory cell access
        array_val = self.run_trec_single(self.expr.value)
        index_val = self.run_trec_single(_get_ast_slice(self.expr))
        self.resmap[0] = mast.Name()
        self.proc('read', self.resmap[0], array_val, index_val)


class UnaryOpHandler(BaseExpressionHandler):
    AST_CLASS = ast.UnaryOp

    def handle(self):
        factory = get_type_map(UNARY_OP_MAP, self.expr.op, 'UnaryOp')
        val = self.run_trec_single(self.expr.operand)
        self.resmap[0] = mast.Name()
        self.pre.append(factory(val, self.resmap[0]))


class BinOpHandler(BaseExpressionHandler):
    AST_CLASS = ast.BinOp

    def handle(self):
        op = get_type_map(BIN_OP_MAP, self.expr.op, 'BinOp')
        left_val = self.run_trec_single(self.expr.left)
        right_val = self.run_trec_single(self.expr.right)
        self.resmap[0] = mast.Name()
        self.proc(f'op {op}', self.resmap[0], left_val, right_val)


class CompareHandler(BaseExpressionHandler):
    AST_CLASS = ast.Compare

    def handle(self):
        end_label = mast.Label()
        self.resmap[0] = mast.Name()
        a_val = self.run_trec_single(self.expr.left)

        for op, comparator in zip(self.expr.ops, self.expr.comparators):
            op = get_type_map(COND_OP_MAP, op, 'Compare')
            b_val = self.run_trec_single(comparator)
            self.proc(f'op {op}', self.resmap[0], a_val, b_val)
            self.jump(end_label, 'equal', self.resmap[0], mast.Literal(False))

            a_val = b_val

        self.pre.append(end_label)


class BoolOpHandler(BaseExpressionHandler):
    AST_CLASS = ast.BoolOp

    def handle(self):
        op, shortcut_condition = get_type_map(
            BOOL_OP_MAP, self.expr.op, 'BoolOp')

        end_label = mast.Label()
        self.resmap[0] = mast.Name()
        val = self.run_trec_single(self.expr.values[0])
        self.proc('set', self.resmap[0], val)
        self.jump(
            end_label, shortcut_condition, val, mast.Literal(False))

        bool_value = mast.Name()

        for value in self.expr.values[1:]:
            val = self.run_trec_single(value)
            self.proc(f'op {op}', bool_value, self.resmap[0], val)
            self.proc('set', self.resmap[0], val)
            self.jump(
                end_label, shortcut_condition, bool_value, mast.Literal(False))

        self.pre.append(end_label)


class IfExpHandler(BaseExpressionHandler):
    AST_CLASS = ast.IfExp

    def handle(self):
        else_label = mast.Label()
        end_label = mast.Label()

        self.resmap[0] = mast.Name()
        cond = self.run_trec_single(self.expr.test)
        self.jump(else_label, 'equal', cond, mast.Literal(False))

        val = self.run_trec_single(self.expr.body)
        self.proc('set', self.resmap[0], val)
        self.jump(end_label, 'always')

        self.pre.append(else_label)

        val = self.run_trec_single(self.expr.orelse)
        self.proc('set', self.resmap[0], val)

        self.pre.append(end_label)


def _create_unary_op(token):
    def fn(self, a):
        self.resmap[0] = mast.Name()
        self.proc(f'op {token}', self.resmap[0], a)
    return fn


def _create_bin_op(token):
    def fn(self, a, b):
        self.resmap[0] = mast.Name()
        self.proc(f'op {token}', self.resmap[0], a, b)
    return fn


_ZERO = mast.Literal(0)


def build_attr_index(method_map):
    patterns = {}
    for name, method in method_map.items():
        nm = tuple(1 if n == '1' else n for n in name.split('__'))
        cur = patterns
        for n in nm:
            if n not in cur:
                cur[n] = {}
            cur = cur[n]
        assert 'method' not in cur
        cur['method'] = method

    def resolve(nm):
        pnames = []
        cur = patterns
        for n in nm:
            if n.name is not None and n.name in cur:
                cur = cur[n.name]
            elif 1 in cur:
                pnames.append(n)
                cur = cur[1]
            else:
                raise IndexError(f'Unresolvable name {".".join(nm)}')
        if 'method' not in cur:
            raise IndexError(f'Unresolvable name {".".join(nm)}')
        return cur['method'], pnames

    return resolve


def no_at_const(n):
    return mast.Name(n.name.lstrip('@'))


def sort_dir_fn(n):
    if not isinstance(n, mast.Name):
        return n
    elif n.name in ('@asc', 'asc'):
        return mast.Literal(1)
    elif n.name in ('@desc', 'desc'):
        return mast.Literal(-1)
    else:
        return n


class CallHandler(BaseExpressionHandler):
    AST_CLASS = ast.Call

    # TODO: support multiple values
    func__M__min = _create_bin_op('min')
    # TODO: support multiple values
    func__M__max = _create_bin_op('max')
    func__M__atan2 = _create_bin_op('atan2')
    func__M__dst = _create_bin_op('dst')
    func__M__noise = _create_bin_op('noise')
    func__M__abs = _create_unary_op('abs')
    func__M__log = _create_unary_op('log')
    func__M__log10 = _create_unary_op('log10')
    func__M__sin = _create_unary_op('sin')
    func__M__cos = _create_unary_op('cos')
    func__M__tan = _create_unary_op('tan')
    func__M__floor = _create_unary_op('floor')
    func__M__ceil = _create_unary_op('ceil')
    func__M__sqrt = _create_unary_op('sqrt')
    func__M__rand = _create_unary_op('rand')

    def func__M__print(self, *args):
        for arg in args:
            self.proc('print', arg)

    def func__1__printFlush(self, target):
        self.proc('printflush', target)

    def func__M__exit(self):
        self.proc('end')

    def func__M__linkCount(self):
        self.resmap[0] = mast.Name()
        self.proc('set', self.resmap[0], mast.Name('@links'))

    def func__M__getLink(self, index):
        self.resmap[0] = mast.Name()
        self.proc('getlink', self.resmap[0], index)

    def func__1__radar(
            self, unit, target1, target2, target3, sort_type, sort_dir):
        self.resmap[0] = mast.Name()
        self.proc(
            'radar',
            no_at_const(target1), no_at_const(target2), no_at_const(target3),
            no_at_const(sort_type), unit,
            sort_dir_fn(sort_dir), self.resmap[0])

    def func__1__sensor(self, unit, prop):
        self.resmap[0] = mast.Name()
        self.proc('sensor', self.resmap[0], unit, prop)

    def func__M__unit__bind(self, utype):
        self.proc('ubind', utype)

    def func__M__unit__radar(
            self, target1, target2, target3, sort_type, sort_dir):
        self.resmap[0] = mast.Name()
        self.proc(
            'uradar',
            no_at_const(target1), no_at_const(target2), no_at_const(target3),
            no_at_const(sort_type), mast.Name('turret1'),
            sort_dir_fn(sort_dir), self.resmap[0])

    def func__M__locate__building(self, block_type, enemy):
        found = self.resmap[0] = mast.Name()
        x = self.resmap[1] = mast.Name()
        y = self.resmap[2] = mast.Name()
        building = self.resmap[3] = mast.Name()
        self.proc(
            'ulocate building', no_at_const(block_type), enemy,
            mast.Name('@copper'),
            x, y, found, building)

    def func__M__locate__ore(self, material):
        found = self.resmap[0] = mast.Name()
        x = self.resmap[1] = mast.Name()
        y = self.resmap[2] = mast.Name()
        self.proc(
            'ulocate ore', mast.Name('core'), mast.Literal(True),
            material, x, y, found, mast.Name())

    def func__M__locate__spawn(self):
        found = self.resmap[0] = mast.Name()
        x = self.resmap[1] = mast.Name()
        y = self.resmap[2] = mast.Name()
        building = self.resmap[3] = mast.Name()
        self.proc(
            'ulocate spawn', mast.Name('core'), mast.Literal(True),
            mast.Name('@copper'), x, y, found, building)

    def func__M__locate__damaged(self):
        found = self.resmap[0] = mast.Name()
        x = self.resmap[1] = mast.Name()
        y = self.resmap[2] = mast.Name()
        building = self.resmap[3] = mast.Name()
        self.proc(
            'ulocate damaged', mast.Name('core'), mast.Literal(True),
            mast.Name('@copper'), x, y, found, building)

    def func__M__draw__clear(self, r, g, b):
        self.proc('draw clear', r, g, b)

    def func__M__draw__color(self, r, g, b, a):
        self.proc('draw color', r, g, b, a)

    def func__M__draw__stroke(self, width):
        self.proc('draw stroke', width)

    def func__M__draw__line(self, x, y, x2, y2):
        self.proc('draw line', x, y, x2, y2)

    def func__M__draw__rect(self, x, y, width, height):
        self.proc('draw rect', x, y, width, height)

    def func__M__draw__lineRect(self, x, y, width, height):
        self.proc('draw lineRect', x, y, width, height)

    def func__M__draw__poly(self, x, y, sides, radius, rotation):
        self.proc('draw poly', x, y, sides, radius, rotation)

    def func__M__draw__linePoly(self, x, y, sides, radius, rotation):
        self.proc('draw linePoly', x, y, sides, radius, rotation)

    def func__M__draw__triangle(self, x, y, x2, y2, x3, y3):
        self.proc('draw triangle', x, y, x2, y2, x3, y3)

    def func__M__draw__image(self, x, y, image, size, rotation):
        self.proc('draw image', x, y, image, size, rotation)

    def func__1__drawFlush(self, target):
        self.proc('drawflush', target)

    def func__1__setEnabled(self, unit, is_enabled):
        self.proc('control enabled', unit, is_enabled)

    def func__1__targetPosition(self, unit, x, y, shoot):
        self.proc('control shoot', unit, x, y, shoot)

    def func__1__targetObject(self, unit, target, shoot):
        self.proc('control shootp', unit, target, shoot)

    def func__1__configure(self, unit, value):
        self.proc('control configure', unit, value)

    def func__M__unit__stop(self):
        self.proc('ucontrol stop', _ZERO, _ZERO, _ZERO, _ZERO, _ZERO)

    def func__M__unit__move(self, x, y):
        self.proc('ucontrol move', x, y, _ZERO, _ZERO, _ZERO)

    def func__M__unit__approach(self, x, y, radius):
        self.proc('ucontrol approach', x, y, radius, _ZERO, _ZERO)

    def func__M__unit__boost(self, value):
        self.proc('ucontrol boost', value, _ZERO, _ZERO, _ZERO, _ZERO)

    def func__M__unit__pathfind(self):
        self.proc('ucontrol pathfind', _ZERO, _ZERO, _ZERO, _ZERO, _ZERO)

    def func__M__unit__targetPosition(self, x, y, shoot):
        self.proc('ucontrol target', x, y, shoot, _ZERO, _ZERO)

    def func__M__unit__targetObject(self, unit, shoot):
        self.proc('ucontrol targetp', unit, shoot, _ZERO, _ZERO, _ZERO)

    def func__M__unit__itemDrop(self, target, amount):
        self.proc('ucontrol itemDrop', target, amount, _ZERO, _ZERO, _ZERO)

    def func__M__unit__itemTake(self, target, material, amount):
        self.proc('ucontrol itemTake', target, material, amount, _ZERO, _ZERO)

    def func__M__unit__payDrop(self):
        self.proc('ucontrol payDrop', _ZERO, _ZERO, _ZERO, _ZERO, _ZERO)

    def func__M__unit__payTake(self, amount):
        self.proc('ucontrol payTake', amount, _ZERO, _ZERO, _ZERO, _ZERO)

    def func__M__unit__mine(self, x, y):
        self.proc('ucontrol mine', x, y, _ZERO, _ZERO, _ZERO)

    def func__M__unit__setFlag(self, value):
        self.proc('ucontrol flag', value, _ZERO, _ZERO, _ZERO, _ZERO)

    def func__M__unit__build(self, x, y, block, rotation, config):
        self.proc('ucontrol build', x, y, block, rotation, config)

    def func__M__unit__getBlock(self, x, y):
        btype = self.resmap[0] = mast.Name()
        unit = self.resmap[1] = mast.Name()
        self.proc('ucontrol getBlock', x, y, btype, unit, _ZERO)

    def func__M__unit__within(self, x, y, radius):
        self.resmap[0] = mast.Name()
        self.proc('ucontrol within', x, y, radius, self.resmap[0], _ZERO)

    def func__1(self, fname, *args):
        fname = fname.name
        if fname not in self.scope:
            raise NameError(f'Undefined function {fname}')
        fdef = self.scope[fname]
        assert isinstance(fdef, utils.FuncDef)
        if len(args) < fdef.n_args:
            raise TypeError(f'Insufficient arguments for function {fname}')

        for fa, sa in zip(fdef.args, args):
            self.proc('set', fa, sa)

        self.proc(
            'op add', fdef.return_addr, mast.Name('@counter'), mast.Literal(1))
        self.jump(fdef.start_label, 'always')
        for k, v in fdef.resmap.items():
            self.resmap[k] = v

    _resolver = staticmethod(build_attr_index({
        k[len('func__'):]: v
        for k, v in vars().items()
        if k.startswith('func__')
    }))

    def resolve_func(self, value):
        if isinstance(value, ast.Name):
            return [self.run_trec_single(value)]
        elif isinstance(value, ast.Attribute):
            return [*self.resolve_func(value.value), mast.Name(value.attr)]
        else:
            raise ValueError(
                'Expressions resulting in functions are not allowed, '
                'only direct calls of named functions: func(1, 2, 3)'
            )

    def handle(self):
        nm = self.resolve_func(self.expr.func)
        method, pre_args = self._resolver(nm)

        if self.expr.keywords:
            raise ValueError('Keyword arguments are not supported')

        arg_vals = [self.run_trec_single(arg) for arg in self.expr.args]
        method(self, *pre_args, *arg_vals)


class AttributeHandler(BaseExpressionHandler):
    AST_CLASS = ast.Attribute

    def prop__M__at__1(self, attr):
        self.resmap[0] = mast.Name(f'@{attr.replace("_", "-")}')

    def prop__M__n__1(self, attr):
        self.resmap[0] = mast.Name(attr)

    def prop__M__sort__1(self, attr):
        if attr == 'asc':
            self.resmap[0] = mast.Literal(1)
        elif attr == 'desc':
            self.resmap[0] = mast.Literal(-1)
        else:
            self.resmap[0] = mast.Name(attr)

    def prop__1__1(self, unit, prop):
        self.resmap[0] = mast.Name()
        self.proc(
            'sensor', self.resmap[0],
            unit, mast.Name(f'@{prop.replace("_", "-")}'))

    def prop__M__unit__1(self, prop):
        self.prop__1__1('@unit', prop)

    def prop__M__at__unit__1(self, prop):
        self.prop__M__unit__1(prop)

    _resolver = staticmethod(build_attr_index({
        k[len('prop__'):]: v
        for k, v in vars().items()
        if k.startswith('prop__')
    }))

    def resolve_value(self, value):
        if isinstance(value, ast.Name):
            return [self.run_trec_single(value)]
        elif isinstance(value, ast.Attribute):
            return [*self.resolve_value(value.value), mast.Name(value.attr)]
        else:
            raise ValueError(
                'Expressions are not allowed before attribute access, '
                'use names of objects directly: Material.copper'
            )

    def handle(self):
        nm = self.resolve_value(self.expr)
        method, pre_args = self._resolver(nm)
        if method.__name__.endswith('__1'):
            pre_args[-1] = pre_args[-1].name
        method(self, *pre_args)


# Statements ===================================


class ExprStatementHandler(BaseExpressionHandler):
    AST_CLASS = ast.Expr

    def handle(self):
        self.run_trec(self.expr.value)


def _check_assignment_to_reserved(name):
    if name in RESERVED_NAMES:
        raise ValueError(f'The name {name} is reserved')


class AssignStatementHandler(BaseExpressionHandler):
    AST_CLASS = ast.Assign

    def _assign(self, targets, values):
        assert len(targets) <= len(values)
        for target, value in zip(targets, values):
            target = self.run_trec_single(target)
            self.proc('set', target, value)

    def named_assign(self, target, value):
        _check_assignment_to_reserved(target.id)
        retvals = self.run_trec(value)
        self._assign([target], retvals)

    def tuple_assign(self, target, value):
        assert all(isinstance(n, ast.Name) for n in target.elts)
        retvals = self.run_trec(value)
        self._assign(target.elts, retvals)

    def memory_assign(self, target, value):
        if not isinstance(target.value, ast.Name):
            raise ValueError(f'Unsupported assignment target {target}')
        _check_assignment_to_reserved(target.value.id)
        index_val = self.run_trec_single(_get_ast_slice(target))
        value_val = self.run_trec_single(value)
        self.proc('write', value_val, mast.Name(target.value.id), index_val)

    TARGET_MAP = {
        ast.Name: named_assign,
        ast.Subscript: memory_assign,
        ast.Tuple: tuple_assign,
    }

    def handle(self):
        if len(self.expr.targets) != 1:
            raise ValueError(
                'Only single target can be used in assignment: a = 3')
        target = self.expr.targets[0]
        if type(target) not in self.TARGET_MAP:
            raise ValueError(f'Unsupported assignment target {target}')
        method = self.TARGET_MAP[type(target)]
        method(self, target, self.expr.value)


class AugAssignStatementHandler(BaseExpressionHandler):
    AST_CLASS = ast.AugAssign

    def named_assign(self, target, op, operand):
        operand_val = self.run_trec_single(operand)
        t = self.run_trec_single(target)
        self.proc(f'op {op}', t, t, operand_val)

    def memory_assign(self, target, op, operand):
        if not isinstance(target.value, ast.Name):
            raise ValueError(f'Unsupported assignment target {target}')
        index_val = self.run_trec_single(_get_ast_slice(target))
        operand_val = self.run_trec_single(operand)
        op_output = mast.Name()
        cell = self.run_trec_single(target.value)
        self.proc('read', op_output, cell, index_val)
        self.proc(f'op {op}', op_output, op_output, operand_val)
        self.proc('write', op_output, cell, index_val)

    TARGET_MAP = {
        ast.Name: named_assign,
        ast.Subscript: memory_assign,
    }

    def handle(self):
        target = self.expr.target
        if type(target) not in self.TARGET_MAP:
            raise ValueError(f'Unsupported assignment target {target}')
        method = self.TARGET_MAP[type(target)]
        op = get_type_map(BIN_OP_MAP, self.expr.op, 'BinOp')
        method(self, target, op, self.expr.value)


class IfStatementHandler(BaseExpressionHandler):
    AST_CLASS = ast.If

    def handle(self):
        end_label = mast.Label()
        else_label = mast.Label() if self.expr.orelse else end_label

        test_val = self.run_trec_single(self.expr.test)
        self.jump(else_label, 'equal', test_val, mast.Literal(False))

        for stmt in self.expr.body:
            self.run_trec(stmt)

        if self.expr.orelse:
            self.jump(end_label, 'always')
            self.pre.append(else_label)

            for stmt in self.expr.orelse:
                self.run_trec(stmt)

        self.pre.append(end_label)


class WhileStatementHandler(BaseExpressionHandler):
    AST_CLASS = ast.While

    def handle(self):
        loop_label = mast.Label()
        end_label = mast.Label()

        self.pre.append(loop_label)
        test_val = self.run_trec_single(self.expr.test)
        self.jump(end_label, 'equal', test_val, mast.Literal(False))

        for stmt in self.expr.body:
            self.run_trec(stmt)

        self.jump(loop_label, 'always')
        self.pre.append(end_label)


class FunctionDefStatementHandler(BaseExpressionHandler):
    AST_CLASS = ast.FunctionDef

    def handle(self):
        if _PY >= (3, 8):
            assert self.expr.args.posonlyargs == []
            assert self.expr.type_comment is None
        assert self.expr.args.vararg is None
        assert self.expr.args.kwonlyargs == []
        assert self.expr.args.kw_defaults == []
        assert self.expr.args.kwarg is None
        assert self.expr.args.defaults == []
        assert self.expr.decorator_list == []
        assert self.expr.returns is None

        fdef = utils.FuncDef(self.expr.name, len(self.expr.args.args))
        self.scope[self.expr.name] = fdef

        end_label = mast.Label()
        self.jump(end_label, 'always')  # skip function body
        self.pre.append(fdef.start_label)

        before, self.pre = self.pre, []

        with self.sub_scope():
            self.scope._current_func = fdef
            for a, n in zip(self.expr.args.args, fdef.args):
                self.scope[a.arg] = n

            for stmt in self.expr.body:
                self.run_trec(stmt)

            fdef.create_return(self)

        self.pre.append(end_label)

        # we have to do this here, since we need to know
        # max size of return tuple
        body, self.pre = self.pre, before
        for rv in fdef.resmap.values():
            self.proc('set', rv, mast.Literal(None))
        self.pre.extend(body)


class ReturnHandler(BaseExpressionHandler):
    AST_CLASS = ast.Return

    def handle(self):
        fdef = self.scope.current_func
        if fdef is None:
            raise ValueError('Returning outside function context')

        if self.expr.value is not None:
            for index, value in enumerate(self.run_trec(self.expr.value)):
                self.resmap[index] = value
                if index not in fdef.resmap:
                    fdef.resmap[index] = mast.Name()
                self.proc('set', fdef.resmap[index], value)

        fdef.create_return(self)


class ImportFromHandler(BaseExpressionHandler):
    AST_CLASS = ast.ImportFrom

    def handle(self):
        pass  # intentionally do nothing


AST_NODE_MAP = {
    subcls.AST_CLASS: subcls
    for subcls in BaseExpressionHandler.__subclasses__()
}


def _create_ast_constant_hack(ast_cls, conv):
    def cons(anode, *args, **kwargs):
        return ConstantHandler(ast.Constant(conv(anode)), *args, **kwargs)
    AST_NODE_MAP[ast_cls] = cons


if _PY < (3, 8):
    _create_ast_constant_hack(ast.Num, lambda expr: expr.n)
    _create_ast_constant_hack(ast.NameConstant, lambda expr: expr.value)
    _create_ast_constant_hack(ast.Str, lambda expr: expr.s)
    _create_ast_constant_hack(ast.Ellipsis, lambda expr: ...)


def transform_expr(expr, scope):
    htype = get_type_map(AST_NODE_MAP, expr, 'expression')
    node = htype(expr, transform_expr, scope)
    node.handle()
    return node.get_results(), node.pre



from dataclasses import dataclass, field
from typing import Dict, List, Optional

from . import mast


@dataclass
class FuncDef:
    name: str
    n_args: int
    args: List[mast.Name] = field(
        repr=False, init=False, default=None)
    resmap: Dict[int, mast.Name] = field(init=False, default_factory=dict)
    start_label: mast.Label = field(
        repr=False, init=False, default_factory=mast.Label)
    return_addr: mast.Name = field(
        repr=False, init=False, default_factory=mast.Name)

    def __post_init__(self):
        self.args = [mast.Name() for _ in range(self.n_args)]

    def create_return(self, handler):
        handler.proc('set', mast.Name('@counter'), self.return_addr)


@dataclass
class Scope:
    _parent_scope: Optional['Scope'] = field(repr=False)
    _current_func: Optional[FuncDef] = field(repr=False, default=None)
    _names: dict = field(init=False, default_factory=dict)

    def __getitem__(self, key):
        if key in self._names:
            return self._names[key]
        if self._parent_scope is not None:
            return self._parent_scope.__getitem__(key)
        raise KeyError

    def __setitem__(self, key, value):
        self._names[key] = value

    def __contains__(self, key):
        if key in self._names:
            return True
        if self._parent_scope is not None:
            return self._parent_scope.__contains__(key)
        return False

    @property
    def current_func(self):
        if self._current_func is not None:
            return self._current_func
        if self._parent_scope is not None:
            return self._parent_scope.current_func
        return None



from . import mast


def _find_all_positions(line, char):
    result = []
    pos = -1
    while True:
        pos = line.find(char, pos + 1)
        if pos >= 0:
            result.append(pos)
        else:
            break
    return result


def _split_cmd(line):
    spc = _find_all_positions(line, ' ')
    quo = _find_all_positions(line, '"')
    for a, b in zip(quo[::2], quo[1::2]):
        spc = [p for p in spc if p < a or p > b]
    spc = [-1] + spc + [len(line)]
    return [line[a + 1:b] for a, b in zip(spc, spc[1:])]


assert _split_cmd('a b c') == ['a', 'b', 'c']
assert _split_cmd('op add a a 1') == ['op', 'add', 'a', 'a', '1']
assert _split_cmd('cmd') == ['cmd']
assert _split_cmd('"a" "b" "c"') == ['"a"', '"b"', '"c"']
assert _split_cmd('"a a" "b b" "c\nc"') == ['"a a"', '"b b"', '"c\nc"']


def _name_or_literal(line):
    if line == 'null':
        return mast.Literal(None)
    if line == 'true':
        return mast.Literal(1.0)
    if line == 'false':
        return mast.Literal(0.0)
    if line[:1] == '"':
        assert line.endswith('"')
        return mast.Literal(line[1:-1])
    try:
        return mast.Literal(float(line))
    except ValueError:
        pass
    return mast.Name(line)


def _parse_cmd(cmd, *args):
    if cmd == 'op':
        return mast.FunctionCall(
            f'op {args[0]}',
            [_name_or_literal(a) for a in args[2:]],
            mast.Name(args[1]),
        )
    elif cmd in ('set', 'read'):
        return mast.FunctionCall(
            cmd,
            [_name_or_literal(a) for a in args[1:]],
            mast.Name(args[0]),
        )
    elif cmd in ('print', 'printflush', 'end', 'write', 'sensor'):
        return mast.ProcedureCall(
            cmd,
            [_name_or_literal(a) for a in args],
        )
    elif cmd == 'jump':
        return mast.Jump(
            int(args[0]),  # label
            args[1],  # condition
            [_name_or_literal(a) for a in args[2:]],
        )
    else:
        raise ValueError(f'Unknown command {cmd} {" ".join(args)}')


def parse(lines):
    result = []
    for line in lines.splitlines():
        line = line.strip()
        if not line:
            continue
        result.append(_parse_cmd(*_split_cmd(line)))
    return result



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['minpiler']

package_data = \
{'': ['*']}

extras_require = \
{":python_version <= '3.6'": ['dataclasses']}

entry_points = \
{'console_scripts': ['minpiler = minpiler.cmdline:run']}

setup(name='minpiler',
      version='0.3.2',
      description='Compile python code to Mindustry microprocessor instructions',
      author='Vitalik Verhovodov',
      author_email='knifeslaughter@gmail.com',
      url='https://github.com/neumond/minpiler',
      packages=packages,
      package_data=package_data,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.6',
     )



from void import (
    M,
    time, allocated_flags,
)


time += 0.5
if time > 360:
    time %= 360

M.unit.bind(M.at.nova)

f = M.unit.flag
if f == 0:
    allocated_flags += 1
    M.unit.setFlag(allocated_flags)
    f = allocated_flags

found, bx, by, building = M.locate.building(M.at.core, False)
if found:
    sx = bx + M.sin(time + f * 30) * 10
    sy = by + M.cos(time + f * 30) * 10

    M.unit.move(sx, sy)
    M.unit.targetPosition(bx, by, True)



import math
import random

from . import mast


class Var:
    __slots__ = ('value', )

    def __init__(self, val):
        self.value = val

    def __repr__(self):
        return f'<Var {repr(self.value)}>'

    @property
    def is_object(self) -> bool:
        v = self.value
        if v is None:
            return True
        if isinstance(v, (int, float)):
            return False
        return True

    @property
    def num_value(self) -> float:
        v = self.value
        if v is None:
            return 0.0
        if v is True:
            return 1.0
        if v is False:
            return 0.0
        if isinstance(v, int):
            return float(v)
        if isinstance(v, float):
            if math.isnan(v) or math.isinf(v):
                return 0.0
            return v
        return 1.0

    @property
    def int_value(self) -> int:
        return int(self.num_value)

    @property
    def printable(self) -> str:
        v = self.value
        if v is None:
            return 'null'
        elif isinstance(v, float):
            return str(v).rstrip('0').rstrip('.')
        elif isinstance(v, dict):
            return v.get('__str__', 'cell')
        elif isinstance(v, str):
            return v.replace(r'\n', '\n')
        else:
            return str(v)


def require_type(val, type_, fn, fallback):
    if isinstance(val, type_):
        return val
    return fallback


def _eq_func(a, b):
    if a.is_object and b.is_object:
        return a.value == b.value
    return a.num_value == b.num_value


def _read_func(cell, index):
    if not isinstance(cell.value, dict):
        return 0.0
    return cell.value.get(index.int_value, 0.0)


def _dst_func(a, b):
    a, b = a.num_value, b.num_value
    return math.sqrt(a * a + b * b)


_DTR = math.pi / 180
_RTD = 180 / math.pi


INST_MAP = {
    'op add': lambda a, b: a.num_value + b.num_value,
    'op sub': lambda a, b: a.num_value - b.num_value,
    'op mul': lambda a, b: a.num_value * b.num_value,
    'op div': lambda a, b: a.num_value / b.num_value,
    'op idiv': lambda a, b: a.num_value // b.num_value,
    'op mod': lambda a, b: a.num_value % b.num_value,
    'op pow': lambda a, b: a.num_value ** b.num_value,
    'op shl': lambda a, b: float(a.int_value << b.int_value),
    'op shr': lambda a, b: float(a.int_value >> b.int_value),
    'op or': lambda a, b: float(a.int_value | b.int_value),
    'op xor': lambda a, b: float(a.int_value ^ b.int_value),
    'op and': lambda a, b: float(a.int_value & b.int_value),
    'op land': lambda a, b: float(bool(a.num_value and b.num_value)),
    'op not': lambda a: float(~a.int_value),
    'op equal': lambda a, b: float(_eq_func(a, b)),
    'op notEqual': lambda a, b: float(not _eq_func(a, b)),
    'op lessThan': lambda a, b: float(a.num_value < b.num_value),
    'op lessThanEq': lambda a, b: float(a.num_value <= b.num_value),
    'op greaterThan': lambda a, b: float(a.num_value > b.num_value),
    'op greaterThanEq': lambda a, b: float(a.num_value >= b.num_value),
    'op min': lambda a, b: min(a.num_value, b.num_value),
    'op max': lambda a, b: max(a.num_value, b.num_value),
    'op atan2': lambda a, b: _RTD * math.atan2(a.num_value, b.num_value),
    'op dst': _dst_func,
    # TODO: op noise
    'op abs': lambda a: abs(a.num_value),
    'op log': lambda a: math.log(a.num_value),
    'op log10': lambda a: math.log10(a.num_value),
    'op sin': lambda a: math.sin(_DTR * a.num_value),
    'op cos': lambda a: math.cos(_DTR * a.num_value),
    'op tan': lambda a: math.tan(_DTR * a.num_value),
    'op floor': lambda a: float(math.floor(a.num_value)),
    'op ceil': lambda a: float(math.ceil(a.num_value)),
    'op sqrt': lambda a: math.sqrt(a.num_value),
    'op rand': lambda a: random.random() * a.num_value,
    'set': lambda a: a.value,
    'read': _read_func,
}
JUMP_CONDS = {
    'equal', 'notEqual',
    'lessThan', 'lessThanEq',
    'greaterThan', 'greaterThanEq',
}


class OutputBuffer:
    def __init__(self):
        self._print_buf = []

    def print(self, text):
        self._print_buf.append(text)

    def printflush(self):
        r = ''.join(self._print_buf)
        self._print_buf.clear()
        return r


def execute_fn_call(ins, fn, args):
    try:
        return fn(*args)
    except (ZeroDivisionError, OverflowError):
        return 0.0


def execute_proc_call(ins, args, raw_args, buf, state):
    if ins.op == 'print':
        buf.print(args[0].printable)
    elif ins.op == 'printflush':
        pass
    elif ins.op == 'write':
        value = args[0].num_value
        cell = args[1].value
        index = args[2].num_value
        cell[index] = value
    elif ins.op == 'sensor':
        if not args[1].is_object or args[1].value is None:
            state.pop(raw_args[0].name, None)
        else:
            state[raw_args[0].name] = args[1].value.get(raw_args[2].name)
    else:
        raise ValueError('Unknown procedure')


def create_arg_evaluator(state):
    def _eval_arg(arg):
        if isinstance(arg, mast.Literal):
            return Var(arg.value)
        elif isinstance(arg, mast.Name):
            assert arg.name is not None
            return Var(state.get(arg.name, None))
        else:
            raise TypeError('Unable to evaluate argument')
    return _eval_arg


def execute(instructions, state: dict, max_steps=5000):
    state = state.copy()
    _eval_arg = create_arg_evaluator(state)

    inst_map = INST_MAP.copy()

    state['@counter'] = 0
    buf = OutputBuffer()

    for _ in range(max_steps):
        ins = instructions[int(state['@counter'])]
        state['@counter'] = int(state['@counter']) + 1
        if isinstance(ins, mast.FunctionCall):
            state[ins.result.name] = execute_fn_call(
                ins,
                inst_map[ins.op],
                list(map(_eval_arg, ins.args)),
            )
        elif isinstance(ins, mast.ProcedureCall):
            if ins.op == 'end':
                break
            else:
                execute_proc_call(
                    ins,
                    list(map(_eval_arg, ins.args)),
                    ins.args,
                    buf,
                    state,
                )
        elif isinstance(ins, mast.Jump):
            apply = False
            if ins.op == 'always':
                apply = True
            elif ins.op in JUMP_CONDS:
                fn = inst_map[f'op {ins.op}']
                apply = fn(*map(_eval_arg, ins.args))
            else:
                raise ValueError('Unknown jump condition')
            if apply:
                state['@counter'] = ins.label
        else:
            raise ValueError('Unknown instruction')
        if state['@counter'] >= len(instructions):
            break
    else:
        raise OverflowError('Too many steps have been taken')
    return buf.printflush()



import pytest

from minpiler import cmdline, emu, mparse, test_utils


PROGRAMS = """
M.print("hello\\n")
M.print(2 + 2, "\\n")
M.print(2 + 2 * 2 + 8 + 6 * 9 * 3, "\\n")
M.print(-5, "\\n")
M.print(+5, "\\n")
M.print(M.max(M.min(2, 8), 3 + 3), "\\n")
M.print("line", 1, 2, M.print(), 3)
--------------
hello
4
176
-5
5
6
line12null3
==============
2 + 6
a = -1
a += 6
if a > 0:
    M.print("a > 0")
else:
    M.print("a <= 0")
--------------
a > 0
==============
M.print(1)
M.exit()
M.print(2)
--------------
1
==============
M.print("Hi ", 2 + M.min(2, 8) * 2)
--------------
Hi 6
==============
cell1[8] = 9
index = 8
cell1[index] += 3
M.print(cell1[index - 2 + 2])
--------------
12
==============
b = 6
M.print(4 <= b <= 8)
M.print(b > 8)
M.print(b == 6)
M.print(99 < b)
M.print(None is None)
--------------
10101
==============
b = 7
M.print(4 and 8 and b, "\\n")
M.print(b and None, "\\n")
M.print(b and None and True, "\\n")
M.print(False and True)
--------------
7
null
null
0
==============
b = 9
M.print(None or 0 or b, "\\n")
M.print(None or 0, "\\n")
M.print(0 or None, "\\n")
M.print(None or 0 or 3, "\\n")
M.print(None or 0 or 3 or 5)
--------------
9
0
null
3
3
==============
M.print(M.abs(-6), ' ', M.sin(90))
--------------
6 1
==============
i = 0
while i < 5:
    M.print(i)
    i += 1
--------------
01234
==============
pos = 10
vel = -20
pos += vel
if pos < 0 and vel < 0:
    vel = -vel
if pos > 100 and vel > 0:
    vel = -vel
M.print(pos, ' ', vel)
--------------
-10 20
==============
M.print(True if 2 > 3 else False)
M.print(True if 2 < 3 else False)
--------------
01
==============
a, b = M.abs(-2), 3 + 6
M.print(a, ' ', b)
--------------
2 9
==============
def fn(a):
    M.print(a)
    return a + 1

b = 3
M.print(fn(b + 4))
M.print("9")
--------------
789
==============
a = 4

def fn(a):
    a = 6
    a += 1
    M.print(a, "\\n")

fn(3)
M.print(a)
--------------
7
4
==============
def balance(container):
    M.print("Object ", container, "\\n")
    M.print("Lead ", container.lead)

balance(container1)
--------------
Object container
Lead 300
==============
"""


@pytest.mark.parametrize('code,result', test_utils.parse_programs(PROGRAMS))
def test_programs(code, result):
    print(cmdline.py_to_mind(code))
    ast = mparse.parse(cmdline.py_to_mind(code))
    assert emu.execute(ast, {
        'cell1': {},
        'container1': {'@lead': 300, '__str__': 'container'},
    }) == result


CODEGEN = """
==============
M.draw.clear(1, 2, 3)
M.draw.color(1, 2, 3, 200)
M.draw.stroke(3)
M.draw.line(20, 20, 50, 50)
M.draw.rect(20, 20, 10, 10)
M.draw.lineRect(20, 20, 10, 10)
M.draw.poly(50, 50, 6, 30, 10)
M.draw.linePoly(50, 50, 6, 30, 10)
M.draw.triangle(20, 20, 60, 60, 30, 10)
M.draw.image(50, 50, M.at.copper, 32, 0)
display1.drawFlush()
--------------
draw clear 1 2 3
draw color 1 2 3 200
draw stroke 3
draw line 20 20 50 50
draw rect 20 20 10 10
draw lineRect 20 20 10 10
draw poly 50 50 6 30 10
draw linePoly 50 50 6 30 10
draw triangle 20 20 60 60 30 10
draw image 50 50 @copper 32 0
drawflush display1
==============
M.print(1)
M.print(b)
M.print('string')
message1.printFlush()
--------------
print 1
print b
print "string"
printflush message1
==============
M.exit()
--------------
end
==============
M.print(M.getLink(4))
--------------
getlink _r1 4
print _r1
==============
conveyor1.setEnabled(True)
duo1.targetPosition(20, 30, True)
duo2.targetObject(enemy1, False)
unloader1.configure(M.at.copper)
--------------
control enabled conveyor1 true
control shoot duo1 20 30 true
control shootp duo2 enemy1 false
control configure unloader1 @copper
==============
M.print(duo1.radar(
    M.at.enemy, M.at.flying, M.at.any,
    M.at.distance, M.at.asc))
--------------
radar enemy flying any distance duo1 1 _r1
print _r1
==============
M.print(duo1.sensor(M.at.health))
M.print(duo1.health)
--------------
sensor _r1 duo1 @health
print _r1
sensor _r2 duo1 @health
print _r2
==============
M.unit.bind(M.at.poly)
M.print(M.unit.radar(
    M.at.enemy, M.at.ground, M.at.any,
    M.at.health, M.at.desc))
--------------
ubind @poly
uradar enemy ground any health turret1 -1 _r1
print _r1
==============
found, x, y, building = M.locate.building(M.at.core, True)
found, x, y = M.locate.ore(M.at.lead)
found, x, y, building = M.locate.spawn()
found, x, y, building = M.locate.damaged()
--------------
ulocate building core true @copper _r1 _r2 _r3 _r4
set found _r3
set x _r1
set y _r2
set building _r4
ulocate ore core true @lead _r5 _r6 _r7 _r8
set found _r7
set x _r5
set y _r6
ulocate spawn core true @copper _r9 _r10 _r11 _r12
set found _r11
set x _r9
set y _r10
set building _r12
ulocate damaged core true @copper _r13 _r14 _r15 _r16
set found _r15
set x _r13
set y _r14
set building _r16
==============
M.unit.stop()
M.unit.move(10, 20)
M.unit.approach(10, 20, 50)
M.unit.boost(True)
M.unit.pathfind()
M.unit.targetPosition(10, 20, True)
M.unit.targetObject(unit, True)
M.unit.itemDrop(unit, 100)
M.unit.itemTake(unit, M.at.lead, 100)
M.unit.payDrop()
M.unit.payTake(10)
M.unit.mine(10, 20)
M.unit.setFlag(5)
M.unit.build(10, 20, M.at.conveyor, 270, M.at.lead)
btype, unit = M.unit.getBlock(10, 20)
result = M.unit.within(10, 20, 50)
--------------
ucontrol stop 0 0 0 0 0
ucontrol move 10 20 0 0 0
ucontrol approach 10 20 50 0 0
ucontrol boost true 0 0 0 0
ucontrol pathfind 0 0 0 0 0
ucontrol target 10 20 true 0 0
ucontrol targetp unit true 0 0 0
ucontrol itemDrop unit 100 0 0 0
ucontrol itemTake unit @lead 100 0 0
ucontrol payDrop 0 0 0 0 0
ucontrol payTake 10 0 0 0 0
ucontrol mine 10 20 0 0 0
ucontrol flag 5 0 0 0 0
ucontrol build 10 20 @conveyor 270 @lead
ucontrol getBlock 10 20 _r1 _r2 0
set btype _r1
set unit _r2
ucontrol within 10 20 50 _r3 0
set result _r3
==============
"""


@pytest.mark.parametrize(
    'pycode,mlogcode', test_utils.parse_programs(CODEGEN))
def test_codegen(pycode, mlogcode):
    assert mlogcode.strip() == cmdline.py_to_mind(pycode).strip()


@pytest.mark.parametrize('pycode', [
    'print = 4',
    'b = print',
    'print + 8',
    'key = GetLink',
    'Draw[8] = True',
])
def test_reserved_name(pycode):
    with pytest.raises(ValueError):
        cmdline.py_to_mind(pycode)



from void import (
    M,
    scatter1, sorter1, switch1, switch2,
    time,
)

sorter1.configure(M.at.copper if switch1.enabled else M.at.lead)

time += 3
if time > 360:
    time %= 360

sx = scatter1.x + M.sin(time) * 10
sy = scatter1.y + M.cos(time) * 10
scatter1.targetPosition(sx, sy, switch2.enabled)
